package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.bipack.Unsigned
import net.sergeych.crypto2.Multikey.AnyKey
import net.sergeych.crypto2.Multikey.Companion.allOf
import net.sergeych.crypto2.Multikey.Companion.allOfMultikeys
import net.sergeych.crypto2.Multikey.Companion.anyOf
import net.sergeych.crypto2.Multikey.Companion.anyOfMultikeys
import net.sergeych.crypto2.Multikey.Companion.invoke
import net.sergeych.crypto2.Multikey.Companion.someOf
import net.sergeych.crypto2.Multikey.Companion.someOfMultikeys

/**
 * Multi-signed key.
 * An arbitrary combination of [VerifyingPublicKey] to implement any multiple keys scenario, like N of M,
 * and logical expression. Sample usage:
 *
 * ```kotlin
 * val k1 = SigningSecretKey.new().verifyingKey
 * val k2 = SigningSecretKey.new().verifyingKey
 * val k3 = SigningSecretKey.new().verifyingKey
 * val k4 = SigningSecretKey.new().verifyingKey
 *
 * val multikey = (k1 or k2) and (k3 or k4)
 *
 * val b: SealedBox = SealedBox.decode(someData)
 *
 * if( b.isSealedBy(multikey) ) {
 *      println("sealed box is properly sealed by a multikey")
 * }
 * ```
 * To build multikeys, use `and` and `or` infix operators against [VerifyingPublicKey], [Multikey], or even
 * [SigningSecretKey] instances, and shortcut methods:
 *
 * - [someOfMultikeys], [someOf] family for `n of M` logic
 * - [anyOfMultikeys], [anyOf], [allOf], and [allOfMultikeys]
 * - [invoke] for a single-key multikey
 * - [AnyKey] when you need effectively match any key, useful when you need a `var` `Multikey`.
 *
 *  __Important__. When serializing, always serialize as root [Multikey] instance to keep
 * it compatible with any combination.
 */
@Serializable
sealed class Multikey {

    /**
     * Check that the [keys] satisfy the condition of this instance
     */
    abstract fun check(keys: Iterable<VerifyingPublicKey>): Boolean

    /**
     * Check that [verifyingKeys] satisfy the multikey condition
     */
    fun check(vararg verifyingKeys: VerifyingPublicKey): Boolean = check(verifyingKeys.asIterable())

    infix fun or(mk: Multikey): Multikey = SomeOf(1, listOf(this, mk))

    infix fun or(k: VerifyingPublicKey) = SomeOf(1, listOf(this, Multikey(k)))

    infix fun or(k: SigningSecretKey) = SomeOf(1, listOf(this, Multikey(k.verifyingKey)))

    infix fun and(mk: Multikey): Multikey = SomeOf(2, listOf(this, mk))

    infix fun and(k: VerifyingPublicKey) = SomeOf(2, listOf(this, Multikey(k)))

    infix fun and(k: SigningSecretKey) = SomeOf(2, listOf(this, Multikey(k.verifyingKey)))

    /**
     * Multikey instance implementing `m of N` logic against [VerifyingPublicKey] set. Do not use
     * it directly, use any [Multikey.someOfMultikeys] functions instead.
     */
    @Serializable
    @SerialName("k")
    class Keys internal constructor(
        @Unsigned
        val requiredMinimum: Int,
        val validKeys: Set<VerifyingPublicKey>,
    ) : Multikey() {
        override fun check(keys: Iterable<VerifyingPublicKey>): Boolean {
            var matches = 0
            for (signer in keys) {
                if (signer in validKeys) {
                    if (++matches >= requiredMinimum) return true
                }
            }
            return false
        }
    }

    /**
     * Multikey instance implementing `m of N` logic against other [Multikey] instances. Do not use
     * it directly, use any [Multikey.someOfMultikeys] functions instead.
     */
    @Serializable
    @SerialName("n")
    class SomeOf internal constructor(
        @Unsigned
        val requiredMinimum: Int,
        val validKeys: List<Multikey>,
    ) : Multikey() {
        override fun check(keys: Iterable<VerifyingPublicKey>): Boolean {
            var matches = 0
            for (k in validKeys) {
                if (k.check(keys)) {
                    if (++matches >= requiredMinimum) return true
                }
            }
            return false
        }
    }

    /**
     * Special `AnyKey`: no restrictions, any key will satisfy this. In the rare case to mark
     * publicly available operations, etc. Please note it is an object, not a class, and can't
     * be instantiated.
     */
    @Serializable
    object AnyKey : Multikey() {
        override fun check(keys: Iterable<VerifyingPublicKey>): Boolean = true
    }


    companion object {

        operator fun invoke(k: SigningSecretKey): Multikey = Keys(1, setOf(k.verifyingKey))
        operator fun invoke(k: VerifyingPublicKey): Multikey = Keys(1, setOf(k))

        /**
         * Create a multikey instance that requires some keys from a list
         */
        fun someOf(requiredMinimum: Int, vararg keys: VerifyingPublicKey): Multikey =
            Keys(requiredMinimum, keys.toSet())

        /**
         * Create a multikey instance that requires some keys from a list
         */
        fun someOfMultikeys(requiredMinimum: Int, vararg keys: Multikey): Multikey =
            SomeOf(requiredMinimum, keys.toList())

        /**
         * Create a multikey instance that requires some keys from a list
         */
        fun someOfMultikeys(requiredMinimum: Int, keys: List<Multikey>): Multikey =
            SomeOf(requiredMinimum, keys)

        /**
         * Create a multikey instance that requires some keys from a list
         */
        fun someOf(requiredMinimum: Int, keys: List<VerifyingPublicKey>): Multikey =
            Keys(requiredMinimum, keys.toSet())

        /**
         * Create a multikey instance that requires any key from a list
         */
        fun anyOf(vararg keys: VerifyingPublicKey): Multikey = someOf(1, *keys)

        /**
         * Create a multikey instance that requires any key from a list
         */
        fun anyOfMultikeys(vararg keys: Multikey): Multikey = someOfMultikeys(1, *keys)

        /**
         * Create a multikey instance that requires any key from a list
         */
        fun anyOfMultikeys(keys: List<Multikey>): Multikey = someOfMultikeys(1, keys)

        /**
         * Create a multikey instance that requires any key from a list
         */
        fun anyOf(keys: List<VerifyingPublicKey>): Multikey = someOf(1, keys)


        /**
         * Create a multikey instance that requires all keys from a list
         */
        fun allOf(vararg keys: VerifyingPublicKey): Multikey = someOf(keys.size, *keys)

        /**
         * Create a multikey instance that requires all keys from a list
         */
        fun allOfMultikeys(vararg keys: Multikey): Multikey = someOfMultikeys(keys.size, *keys)

        /**
         * Create a multikey instance that requires all keys from a list
         */
        fun allOfMultikeys(keys: List<Multikey>): Multikey = someOfMultikeys(keys.size, keys)

        /**
         * Create a multikey instance that requires all keys from a list
         */
        fun allOf(keys: List<VerifyingPublicKey>): Multikey = someOf(keys.size, keys)


    }
}