package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.decodeFromBipack
import net.sergeych.crypto2.VerifyingPublicKey.Companion.toString
import net.sergeych.mp_tools.decodeBase64Url

/**
 * The public for public-key encryption. It encrypts messages that can only be decrypted with corresponding
 * [SecretKey].
 */
@Serializable
@SerialName("encp")
class PublicKey(override val keyBytes: UByteArray) : UniversalKey(), EncryptingKey {

    override val magic: KeysmagicNumber = KeysmagicNumber.defaultAssymmetric
    @Transient
    override val label: String = "pub"

    /**
     * Create an anonymous message that could be decrypted only with the [SecretKey] that corresponds this.
     * Anonymous message uses one-time secret key, the public part of which is included into the
     * [Asymmetric.Message], so the sender could not be identified.
     *
     * __Anonymous encryption is much slower__ as it generates new keys every time, use [encryptMessage]
     * when possible
     *
     * The authentication is used despite the anonymity, and the fact of the successful decryption
     * proves that the message was not altered after creation.
     */
    fun encryptAnonymousMessage(plainData: UByteArray, randomFill: IntRange? = null): Asymmetric.Message =
        encryptMessage(plainData, randomFill = randomFill)

    /**
     * Anonymous encryption, see [encryptAnonymousMessage], to binary data. Sender could not be identified.
     */
    @Suppress("unused")
    override fun encrypt(plainData: UByteArray, randomFill: IntRange?): UByteArray =
        encryptMessage(plainData, randomFill = randomFill).encoded

    override fun encryptWithNonce(plainData: UByteArray, nonce: UByteArray, randomFill: IntRange?): UByteArray =
        encryptMessage(plainData, nonce = nonce, randomFill = randomFill).encoded

    override val nonceBytesLength: Int = Asymmetric.nonceBytesLength

    /**
     * Universal public-key encryption. Note that message authenticity is guaranteed if the decryption is successful
     * whether [senderKey] is provider, the latter only allow to positively identify the sender.
     *
     * @param plainData data to encrypt
     * @param nonce allows specifying exact nonce, default to random (safe)
     * @param senderKey key to authenticate sending party. It is safe and much faster to specify it,
     *      otherwise an anonymous key will be created for each encryption, also safe and anonymous, but slow.
     */
    fun encryptMessage(
        plainData: UByteArray,
        nonce: UByteArray = randomNonce(),
        senderKey: SecretKey = newSecretKey(),
        randomFill: IntRange? = null,
    ) = Asymmetric.createMessage(senderKey, this, WithFill.encode(plainData, randomFill), nonce)

    /**
     * Encrypt message using the specified secret key as sender authentication. Recipient, the party having
     * [SecretKey] corresponding to this one, will be able to decrypt the message and be sure that [senderKey]
     * was the author and the message was not altered.
     */
    fun encryptMessage(
        plainData: UByteArray,
        senderKey: SecretKey,
        randomFill: IntRange? = null,
    ): Asymmetric.Message =
        Asymmetric.createMessage(senderKey, this, WithFill.encode(plainData, randomFill))

    override val id by lazy { KeyId(magic, keyBytes, null, true) }

    companion object {
        /**
         * Parse any known public key text representation, including what [toString] return (for public keys it is
         * possible)
         * @throws IllegalArgumentException the public key isn't recognized
         */
        fun parse(text: String): PublicKey {
            val s = text.trim()

            fun parseId(t: String): PublicKey{
                val id = BinaryId.restoreFromString(t)
                if( id.magic != KeysmagicNumber.defaultAssymmetric.ordinal)
                    throw IllegalArgumentException("invalid magick ${id.magic} for PublicKey")
                return id.asPublicKey
            }

            // 🗝sig#I1po9Y2I7p2aOxeh4nFyGPm3e0YunBEu1Mo-PmIqP84Evg
            return when {
                s.startsWith("\uD83D\uDDDDpub#") -> parseId(s.drop(6))
                s.startsWith("pub#") -> parseId(s.drop(4))
                s.startsWith("#") -> parseId(s.drop(1))
                else -> {
                    // consider it is serialized key in base64 format
                    val data = s.decodeBase64Url().asUByteArray()
                    if (data.size == 32)
                        PublicKey(data)
                    else {
                        runCatching { data.decodeFromBipack<PublicKey>() }.getOrNull()
                            ?: kotlin.runCatching { data.decodeFromBipack<UniversalKey>() as PublicKey }
                                .getOrElse { throw IllegalArgumentException("can't parse verifying key") }
                    }
                }
            }
        }
    }
}