package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient

/**
 * Combination of private/secret keys suitable for both decryption and signing.
 *
 * It contains two cryptographically independent keys to raise security to a maximum.
 * Any converted keys poses a threat while technically possible so we avoid it.
 */
@Serializable
@SerialName("uprv")
class UniversalPrivateKey(
    val signingKey: SigningSecretKey,
    val decryptingKey: SecretKey
) : UniversalKey(), DecryptingKey by decryptingKey, SigningKey by signingKey {

    override val keyBytes by lazy { signingKey.keyBytes + decryptingKey.keyBytes }

    @Transient
    override val magic = KeysmagicNumber.defaultUniversalPrivate

    /**
     * Important! Private key combines signing and decrypting keys, but uses
     * it of the decrypting one to be used in keyring.
     */
    @Transient
    override val id: KeyId = decryptingKey.id

    /**
     * Corresponding public key able to verify amd encrypt data created by this
     * private key.
     */
    val publicKey by lazy {
        UniversalPublicKey(signingKey.verifyingKey, decryptingKey.publicKey)
    }

    companion object {
        /**
         * Generate 2 new random keys (4 key pairs under the hood) to securely signd and
         * decrypt data.
         */
        fun new() = UniversalPrivateKey(SigningSecretKey.new(), SecretKey.new())
    }
}