package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.pwhash.*
import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import net.sergeych.bipack.Unsigned
import net.sergeych.crypto2.KDF.Argon.Alg.V2id_13
import net.sergeych.crypto2.KDF.Complexity.*
import net.sergeych.crypto2.PBKD.Params

@Serializable
sealed class KDF {

    /**
     * KDF complexity: how many resources (CPY and RAM) it requires to transform a string to a key.
     *
     * Constants started with `Fixed...`, e.g., [FixedLow], [FixedMedium] and [FixedHigher]
     * are first clas constants not to be changed in future but can become obsolete (too simple) with the
     * development of the computer hardware. Other constants can change in future releases, so it is critical
     * to serialize [KDF] or [PBKD.Params] generated with such constants and restore them before
     * deriving keys.
     */
    enum class Complexity {
        FixedLow,
        Interactive,
        FixedMedium,
        Moderate,
        FixedHigher,
        Sensitive,
        ;

        /**
         * Create [KDF] of the corresponding strength suitable to derive [numberOfKeys] symmetric keys.
         *
         * Random salt of proper size is used
         */
        fun kdfForSize(numberOfKeys: Int): KDF = creteDefault(SymmetricKey.keyLength * numberOfKeys, this)

        /**
         * Derive multiple keys from the password. Derivation params will be included in the key ids, see
         * [SymmetricKey.id] as [KeyId.kdp].

         * Random salt of proper size is used
         *
         * ___Important: symmetric keys do not save key ids___. _Container do it, so it is possible to re-derive
         * key to open the container, but in many cases you might need to save [KeyId.kdp] separately_.
         * Having no [PBKD.Params] it would not be possible to recreate the key, as complexity defaults tend
         * to change with time.
         */
        @Suppress("unused")
        fun deriveMultiple(password: String, count: Int): List<SymmetricKey> =
            kdfForSize(count).deriveMultipleKeys(password, count)

        /**
         * Derive single key from password, same as [deriveMultiple] with count=1.
         */
        fun derive(password: String): SymmetricKey = deriveMultiple(password, 1).first()
    }

    /**
     * Derive a single key from the password, same as [deriveMultipleKeys] with `count==1`
     */
    abstract fun deriveKey(password: String): UByteArray

    /**
     * Derive keys from lower part of bytes derived from the password. E.g., if generated size is longer than
     * required to create [count] keys, the first bytes will be used. It let use rest bytes for other purposes.
     */
    fun deriveMultipleKeys(password: String, count: Int): List<SymmetricKey> {
        val bytes = deriveKey(password)
        val ks = SymmetricKey.keyLength
        check(ks * count <= bytes.size) { "KDF is too short for $count keys: ${bytes.size} we need ${ks * count}" }
        return (0..<count).map {
            val base = it * ks
            Params(this, base, ks).deriveKey(password)
        }
    }

    @Serializable
    @SerialName("argon")
    data class Argon(
        val algorithm: Alg,
        @Unsigned
        val instructionsComplexity: ULong,
        @Unsigned
        val memComplexity: Int,
        val salt: UByteArray,
        @Unsigned
        val keySize: Int,
    ) : KDF(), Comparable<Argon> {


        /**
         * Very abstract strength comparison. If a1 > a2 it generally means that its complexity, and, hopefully,
         * strength is higher.
         */
        override fun compareTo(other: Argon): Int {
            var d = algorithm.ordinal.compareTo(other.algorithm.ordinal)
            if (d != 0) return d
            d = instructionsComplexity.compareTo(other.instructionsComplexity)
            if (d != 0) return d
            d = memComplexity.compareTo(other.memComplexity)
            if (d != 0) return d
            d = keySize
            if (d != 0) return d
            return 0
        }

        enum class Alg(val code: Int) {
            @Suppress("EnumEntryName", "unused")
            V2i_13(crypto_pwhash_argon2i_ALG_ARGON2I13),
            V2id_13(crypto_pwhash_argon2id_ALG_ARGON2ID13),
            ;

            companion object {
                val default = V2id_13
            }
        }

        override fun deriveKey(password: String): UByteArray =
            PasswordHash.pwhash(keySize, password, salt, instructionsComplexity, memComplexity, algorithm.code)

        override fun equals(other: Any?): Boolean {
            if (this === other) return true
            if (other !is Argon) return false

            if (algorithm != other.algorithm) return false
            if (instructionsComplexity != other.instructionsComplexity) return false
            if (memComplexity != other.memComplexity) return false
            if (keySize != other.keySize) return false
            return salt contentEquals other.salt
        }

        override fun hashCode(): Int {
            var result = algorithm.hashCode()
            result = 31 * result + instructionsComplexity.hashCode()
            result = 31 * result + memComplexity
            result = 31 * result + salt.contentHashCode()
            result = 31 * result + keySize
            return result
        }


        companion object {
            val saltSize: Int = crypto_pwhash_SALTBYTES
            val minKeySize: Int = crypto_pwhash_BYTES_MIN.toInt()

            fun randomSalt() = randomUBytes(saltSize)

            /**
             * Create a deterministic salt suitable fot this KDF from a given text.
             *
             * We recommend to use random salts stored, and [KeyId] of password-generated keys already
             * do it for you. Use this method only if you can't store [KeyId] or salt; it is generally less secure:
             * knowing the base text it is possible to understand, for example, that the same derived password was used
             * more than once (random salt makes it impossible).
             */
            @Suppress("unused")
            fun deriveSaltFromString(text: String) = Hash.Blake2b.deriveSalt(text, saltSize)

            fun create(complexity: Complexity, salt: UByteArray, keySize: Int): Argon {
                require(salt.size == saltSize) { "The salt size should be $saltSize" }
                require(keySize > minKeySize) { "The key size should be at least $keySize bytes" }
                return when (complexity) {
                    FixedLow -> Argon(
                        V2id_13,
                        2UL,
                        67108864,
                        salt, keySize
                    )

                    Interactive -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_INTERACTIVE.toULong(),
                        crypto_pwhash_MEMLIMIT_INTERACTIVE,
                        salt, keySize
                    )

                    FixedMedium -> Argon(
                        V2id_13,
                        3UL,
                        268435456,
                        salt, keySize
                    )

                    Moderate -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_MODERATE,
                        crypto_pwhash_MEMLIMIT_MODERATE,
                        salt, keySize
                    )

                    Sensitive -> Argon(
                        Alg.default,
                        crypto_pwhash_OPSLIMIT_SENSITIVE,
                        crypto_pwhash_MEMLIMIT_SENSITIVE,
                        salt, keySize
                    )

                    FixedHigher -> Argon(
                        V2id_13,
                        4UL,
                        1073741824,
                        salt, keySize
                    )

                }
            }
        }
    }

    companion object {
        fun creteDefault(keySize: Int, complexity: Complexity, salt: UByteArray = Argon.randomSalt()): KDF {
            return Argon.create(complexity, salt, keySize)
        }

    }

    data class Instance(val kdf: KDF, val password: String)

}

